import os
import logging
from mkdocs.config import config_options
from mkdocs.plugins import BasePlugin
from mkdocs.exceptions import PluginError
from jinja2 import Environment, FileSystemLoader, select_autoescape
from .util import Util
from pathlib import Path
from datetime import datetime, time
import re
import os

DIR_PATH = Path(os.path.dirname(os.path.realpath(__file__)))
BLOG_PAGE_PATTERN = re.compile(
    r"\{\{\s*blog_content\s*\}\}", flags=re.IGNORECASE)
TAG_PAGE_PATTERN = re.compile(
    r"\{\{\s*tag_content\s*\}\}", flags=re.IGNORECASE)
PLACEHOLDER = "{{ PLACEHOLDER }}"
THEMES = ["card", "button"]
file = open(DIR_PATH / "templates" / "pagination.js")
SCRIPTS = "<script>" + file.read() + "</script>"
del file

logger = logging.getLogger("mkdocs.plugins")


class BloggingPlugin(BasePlugin):
    """
    Mkdocs plugin to add blogging functionality
    to mkdocs site.
    """
    config_scheme = (
        ("dirs", config_options.Type(list, default=[])),
        ("size", config_options.Type(int, default=10)),
        ("sort", config_options.Type(
            dict, default={"from": "new", "by": "creation"})),
        ("meta_time_format", config_options.Type(str, default=None)),
        ("locale", config_options.Type(str, default=None)),
        ("paging", config_options.Type(bool, default=True)),
        ("show_total", config_options.Type(bool, default=True)),
        ("template", config_options.Type(str, default=None)),
        ("theme", config_options.Type(dict, default=None)),
        ("features", config_options.Type(dict, default={})),
        ("time_format", config_options.Type(str, default=None)),
    )

    blog_pages = []

    site_url = ""

    # Config
    docs_dirs = []
    size = 0
    sort = {}
    meta_time_format = None
    time_format = None
    locale = None
    paging = True
    show_total = True
    template_file = None
    theme = None
    features = []

    # Blog page
    blog_html = None

    # Tags
    tags_index_template = None
    tags_template = None
    tags = {}
    tags_page_html = None
    tags_index_url = ""

    util = Util()

    def on_serve(self, server, config, builder):
        self.get_template(config)

        if self.template_file:
            # Watch the template file for live reload
            server.watch(self.template_file)

        return server

    def on_config(self, config):
        # Abort with error with 'navigation.instant' feature on
        # because paging won't work with it.
        mkdocs_theme = config.get("theme")
        if mkdocs_theme and "features" in mkdocs_theme and \
                "navigation.instant" in mkdocs_theme["features"] and self.paging:
            raise PluginError("[blogging-plugin] Feature 'navigation.instant' "
                              "cannot be enabled with option 'paging' on.")

        self.site_url = config.get("site_url")

        if not self.template_file:
            self.get_template(config)

        self.size = self.config.get("size")
        self.docs_dirs = [Path(path) for path in self.config.get("dirs")]
        self.paging = self.config.get("paging")
        self.sort = self.config.get("sort")
        self.meta_time_format = self.config.get("meta_time_format")
        self.show_total = self.config.get("show_total")
        self.theme = self.config.get("theme")
        self.features = self.config.get("features")
        self.time_format = self.config.get("time_format")

        if "from" not in self.sort:
            self.sort["from"] = "new"
        if "by" not in self.sort:
            self.sort["by"] = "creation"

        if self.theme:
            if self.template_file:
                logger.warning(
                    "[blogging-plugin] Custom template has higher priority than "
                    "predefined themes"
                )
            elif "name" not in self.theme:
                logger.warning(
                    f"[blogging-plugin] Theme name not found. Using default theme..."
                )
                self.theme = None
            elif self.theme["name"] not in THEMES:
                logger.warning(
                    f"[blogging-plugin] Theme '{self.theme['name']}' not found. "
                    "Using default theme..."
                )
                self.theme = None

        search_paths = [DIR_PATH / "templates"]
        if self.template_file:
            search_paths.append(self.template_file.parents[0])

        env = Environment(
            loader=FileSystemLoader(search_paths),
            autoescape=select_autoescape()
        )

        self.template = env.get_template(
            self.template_file.name if self.template_file else
            (f"blog-{self.theme['name']}-theme.html" if self.theme else "blog.html")
        )

        # TODO: Custom template
        self.tags_index_template = env.get_template("blog-tags-index.html")
        self.tags_template = env.get_template("blog-tags.html")

        if "tags" in self.features:
            index_path = self.features["tags"].get("index_page")
            if index_path:
                index_path = Path(index_path)
                self.tags_index_url = self.site_url + \
                    (index_path.parents[0] / index_path.stem).as_posix()

        if self.config.get("locale"):
            self.locale = self.config.get("locale")
        else:
            self.locale = config.get("locale")

        # Remove all pages to adapt live reload
        self.blog_pages = []
        self.tags = {}

    def on_page_markdown(self, markdown, page, config, files):
        if "tags" in self.features and "tags" in page.meta:
            tags = page.meta["tags"]
            page = self.with_timestamp(page)
            if isinstance(tags, list):
                for tag in tags:
                    if tag not in self.tags:
                        self.tags[tag] = [page]
                    else:
                        self.tags[tag].append(page)
            else:
                logger.warning(
                    f"[blogging-plugin] Tags entry '{tags}' is not a list. "
                    "Skipping..."
                )

            # Insert tags into original page
            insert = self.features["tags"].get("insert")
            if insert:
                tags_html = "\n" + self.tags_template.render(tags=page.meta["tags"],
                                                             index_url=self.tags_index_url).strip() + "\n"
                if insert == "bottom":
                    markdown = markdown + "\n<br/>\n" + tags_html
                else:
                    markdown = tags_html + markdown

        return markdown

    def on_page_content(self, html, page, config, files):
        """
        Add meta information about creation date after the html has
        been generated, the time when the meta from markdown file
        has already been added into the page instance.
        """

        if not self.docs_dirs:
            return

        if "exclude_from_blog" in page.meta and page.meta["exclude_from_blog"]:
            return

        file_path = Path(page.file.src_path)
        for dir in self.docs_dirs:
            dir_path = Path(dir)
            if dir_path in file_path.parents:
                self.blog_pages.append(self.with_timestamp(page))
                break

    def on_post_page(self, output, page, config):
        if self.docs_dirs and self.blog_pages:
            if not self.blog_html:
                self.blog_html = self.generate_html(
                    self.sorted_pages(self.blog_pages))

            result = BLOG_PAGE_PATTERN.subn(
                self.blog_html,
                output,
            )

            # There are matches
            if result[1]:
                """
                Add js script to the end of the document to manipulate paging
                bahaviours.
                """
                output = result[0] + SCRIPTS

        if "tags" in self.features and self.tags:
            tag_names = [tag for tag in self.tags]
            sorted_entries = {tag: self.sorted_pages(
                self.tags[tag]) for tag in self.tags}
            if not self.tags_page_html:
                self.tags_page_html = self.tags_index_template.render(
                    tags=tag_names, entries=sorted_entries,
                    index_url=self.tags_index_url)
            output = TAG_PAGE_PATTERN.sub(self.tags_page_html, output)

        return output

    def get_template(self, config):
        if self.config.get("template"):
            root_url = Path(config.get("config_file_path")).parents[0]
            self.template_file = root_url / self.config.get("template")

    def generate_html(self, pages) -> str:
        blog_pages = sorted(pages,
                            key=lambda page: page.meta["git-timestamp"],
                            reverse=self.sort["from"] == "new"
                            )
        theme_options = self.theme.get("options") if self.theme else []
        return self.template.render(
            pages=blog_pages, page_size=self.size,
            paging=self.paging, is_revision=self.sort["by"] == "revision",
            show_total=self.show_total, theme_options=theme_options,
            index_url=self.tags_index_url, show_tags="tags" in self.features
        )

    def sorted_pages(self, pages):
        return sorted(pages,
                      key=lambda page: page.meta["git-timestamp"],
                      reverse=self.sort["from"] == "new")

    def with_timestamp(self, page):
        timestamp = None
        if self.meta_time_format:
            if "time" in page.meta:
                timestamp = datetime.strptime(
                    page.meta["time"], self.meta_time_format).timestamp()
            elif "date" in page.meta:
                timestamp = datetime.strptime(
                    page.meta["date"], self.meta_time_format).timestamp()
        if not timestamp:
            timestamp = self.util.get_git_commit_timestamp(
                page.file.abs_src_path, is_first_commit=self.sort["by"] != "revision")
        page.meta["git-timestamp"] = timestamp
        page.meta["localized-time"] = self.util.get_localized_date(
            timestamp, False, format=self.time_format, _locale=self.locale)

        return page



"""
MIT License

Copyright (c) 2018 Terry Zhao
Copyright (c) 2019 Tim Vink
Copyright (c) 2021 Liang Yesheng

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

"""Utility class for mkdocs plugin."""
import logging
import os
import time
from git import (
    Git,
    Repo,
    GitCommandError,
    GitCommandNotFound,
    InvalidGitRepositoryError,
    NoSuchPathError,
)

from babel.dates import format_date, format_datetime
import locale
from datetime import datetime

logger = logging.getLogger("mkdocs.plugins")

class Util:
    """Utility class.

    This helps find git and calculate relevant dates.
    """

    def __init__(self):
        """Initialize utility class."""
        self.repo_cache = {}

    def _get_repo(self, path: str) -> Git:
        if not os.path.isdir(path):
            path = os.path.dirname(path)

        if path not in self.repo_cache:
            self.repo_cache[path] = Repo(path, search_parent_directories=True).git

        return self.repo_cache[path]


    def get_git_commit_timestamp(
            self,
            path: str,
            is_first_commit: bool = False
    ) -> int:
        """
        Get a list of commit dates in unix timestamp, starts with the most recent commit.

        Args:
            is_first_commit (bool): if true, get the timestamp of the first commit,
                                    else, get that of the most recent commit.
            path (str): Location of a markdown file that is part of a Git repository.
            is_first_commit (bool): retrieve commit timestamp when file was created.

        Returns:
            int: commit date in unix timestamp, starts with the most recent commit.
        """
        commit_timestamp = ""

        # perform git log operation
        try:
            # Retrieve author date in UNIX format (%at)
            # https://git-scm.com/docs/git-log#Documentation/git-log.txt-ematem
            # https://git-scm.com/docs/git-log#Documentation/git-log.txt---diff-filterACDMRTUXB82308203
            realpath = os.path.realpath(path)
            git = self._get_repo(realpath)
            if is_first_commit:
                # diff_filter="A" will select the commit that created the file
                commit_timestamp = git.log(
                    realpath, date="short", format="%at", diff_filter="A"
                )
                # A file can be created multiple times, through a file renamed. 
                # Commits are ordered with most recent commit first
                # Get the oldest commit only
                if commit_timestamp != "":
                    commit_timestamp = commit_timestamp.split()[-1]
            else:
                commit_timestamp = git.log(
                    realpath, date="short", format="%at", n=1
                )
        except (InvalidGitRepositoryError, NoSuchPathError) as err:
            logger.warning(
                "[blogging-plugin] Unable to find a git directory and/or git is not installed."
                " Falling back to build date."
            )
            commit_timestamp = time.time()
        except GitCommandError as err:
            logger.warning(
                "[blogging-plugin] Unable to read git logs of '%s'. Is git log readable?"
                " Falling back to build date."
                % path
            )
            commit_timestamp = time.time()
        except GitCommandNotFound as err:
            logger.warning(
                "[blogging-plugin] Unable to perform command: 'git log'. Is git installed?"
                " Falling back to build date."
            )
            commit_timestamp = time.time()

        # create timestamp
        if commit_timestamp == "":
            commit_timestamp = time.time()
            logger.warning(
                "[blogging-plugin] '%s' has no git logs, using current timestamp"
                % path
            )

        return int(commit_timestamp)

    @staticmethod
    def get_localized_date(timestamp: float, day_only: bool, format: str=None, _locale: str=None) -> str:
        time = datetime.fromtimestamp(timestamp)
        if format:
            return datetime.strftime(time, format)
        else:
            if not _locale:
                _locale = locale.getdefaultlocale()[0]

            if day_only:
                return format_date(time.date(), format="short", locale=_locale)
            else:
                return format_datetime(time, format="short", locale=_locale)


from setuptools import find_packages, setup

file = open("README.md", "r")
LONG_DESCRIPTION = file.read()
file.close()

file = open("requirements.txt", "r")
DEPENDENCIES = file.readlines()
file.close()

del file

setup(
    name="mkdocs-blogging-plugin",
    version="1.2.1",
    description="Mkdocs plugin that generates a blog index page sorted by creation date.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords="mkdocs blog plugin",
    project_urls={
        "Source": "https://github.com/liang2kl/mkdocs-blogging-plugin"
    },
    author="Liang Yesheang",
    author_email="liang2kl@outlook.com",
    include_package_data=True,
    license="MIT",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=DEPENDENCIES,
    packages=find_packages(),
    entry_points={
        "mkdocs.plugins": [
            "blogging = mkdocs_blogging_plugin.plugin:BloggingPlugin"
        ]
    }
)
